from binascii import hexlify

from binary.unsigned_integer.reader import read_bit8

from crypto.transactions.deserializers.base import BaseDeserializer


class MultiSignatureRegistrationDeserializer(BaseDeserializer):

    def deserialize(self):
        starting_position = int(self.asset_offset / 2)

        self.transaction.asset = {
            'multisignature': {
                'min': read_bit8(self.serialized, starting_position) & 0xff,
                'lifetime': read_bit8(self.serialized, starting_position + 2) & 0xff,
                'keysgroup': []
            }
        }

        count = read_bit8(self.serialized, starting_position + 3) & 0xff
        for index in range(count):
            index_start = int(self.asset_offset) + 6
            if (index > 0):
                index_start += index * 66
            signature = hexlify(self.serialized)[index_start:index_start + 66].decode()
            self.transaction.asset['multisignature']['keysgroup'].append(signature)

        self.transaction.parse_signatures(
            hexlify(self.serialized),
            self.asset_offset + 6 + (count * 66)
        )
        return self.transaction



from crypto.transactions.serializers.base import BaseSerializer


class DelegateResignationSerializer(BaseSerializer):
    """Serializer handling delegate resignation data
    """

    def serialize(self):
        return self.bytes_data



from crypto.constants import TRANSACTION_VOTE
from crypto.identity.address import address_from_passphrase
from crypto.transactions.builder.base import BaseTransactionBuilder


class Vote(BaseTransactionBuilder):

    transaction_type = TRANSACTION_VOTE

    def __init__(self, vote, fee=None):
        """Create a second signature registration transaction

        Args:
            vote (str): address of a delegate you want to vote
            fee (int, optional): fee used for the transaction (default is already set)
        """
        super().__init__()
        self.transaction.asset['votes'] = [vote]
        if fee:
            self.transaction.fee = fee

    def sign(self, passphrase):
        self.transaction.recipientId = address_from_passphrase(passphrase)
        super().sign(passphrase)



from binascii import unhexlify

from binary.unsigned_integer.writer import write_bit8

from crypto.transactions.serializers.base import BaseSerializer


class VoteSerializer(BaseSerializer):
    """Serializer handling vote data
    """

    def serialize(self):
        vote_bytes = []
        for vote in self.transaction['asset']['votes']:
            if vote.startswith('+'):
                vote_bytes.append('01{}'.format(vote[1::]))
            else:
                vote_bytes.append('00{}'.format(vote[1::]))
        self.bytes_data += write_bit8(len(self.transaction['asset']['votes']))
        self.bytes_data += unhexlify(''.join(vote_bytes))
        return self.bytes_data



from crypto.constants import TRANSACTION_TRANSFER
from crypto.transactions.builder.base import BaseTransactionBuilder


class Transfer(BaseTransactionBuilder):

    transaction_type = TRANSACTION_TRANSFER

    def __init__(self, recipientId, amount, vendorField=None, fee=None):
        """Create a transfer transaction

        Args:
            recipientId (str): address to which you want to send coins
            amount (int): amount of coins you want to transfer
            vendorField (str): value for the vendor field aka smartbridge
            fee (int, optional): fee used for the transaction (default is already set)
        """
        super().__init__()
        self.transaction.recipientId = recipientId
        self.transaction.amount = amount
        self.transaction.vendorField = vendorField.encode() if vendorField else None
        if fee:
            self.transaction.fee = fee



import hashlib

from base58 import b58encode_check

from binary.unsigned_integer.writer import write_bit8

from crypto.configuration.network import get_network


def wif_from_passphrase(passphrase, network_wif=None):
    """Get wif from passphrase

    Args:
        passphrase (bytes):
        network_wif (int, optional):

    Returns:
        string: wif
    """
    if not network_wif:
        network = get_network()
        network_wif = network['wif']

    private_key = hashlib.sha256(passphrase.encode())
    seed = write_bit8(network_wif) + private_key.digest() + write_bit8(0x01)
    return b58encode_check(seed).decode()



from binascii import hexlify
from hashlib import sha256

from coincurve import PrivateKey as PvtKey


class PrivateKey(object):

    def __init__(self, private_key):
        self.private_key = PvtKey.from_hex(private_key)
        self.public_key = hexlify(self.private_key.public_key.format()).decode()

    def sign(self, message):
        """Sign a message with this private key object

        Args:
            message (bytes): bytes data you want to sign

        Returns:
            str: signature of the signed message
        """
        signature = self.private_key.sign(message)
        return hexlify(signature).decode()

    def to_hex(self):
        """Returns a private key in hex format

        Returns:
            str: private key in hex format
        """
        return self.private_key.to_hex()

    @classmethod
    def from_passphrase(cls, passphrase):
        """Create PrivateKey object from a given passphrase

        Args:
            passphrase (str):

        Returns:
            PrivateKey: Private key object
        """
        private_key = sha256(passphrase.encode()).hexdigest()
        return cls(private_key)

    @classmethod
    def from_hex(cls, private_key):
        """Create PrivateKey object from a given hex private key

        Args:
            private_key (str):

        Returns:
            PrivateKey: Private key object
        """
        return cls(private_key)



from binascii import unhexlify

from crypto.transactions.serializers.base import BaseSerializer


class SecondSignatureRegistrationSerializer(BaseSerializer):
    """Serializer handling registration of the seconds signature
    """

    def serialize(self):
        self.bytes_data += unhexlify(self.transaction['asset']['signature']['publicKey'].encode())
        return self.bytes_data



from datetime import datetime

from crypto.configuration.network import get_network


def get_time():
    """Get the time difference between now and network start.

    Returns:
        int: difference in seconds
    """
    now = datetime.utcnow()
    network = get_network()
    seconds = int((now - network['epoch']).total_seconds())
    return seconds


def get_epoch():
    network = get_network()
    return network['epoch']



class BaseSerializer(object):

    transaction = None
    bytes_data = None

    def __init__(self, transaction, bytes_data=bytes()):
        self.transaction = transaction
        self.bytes_data = bytes_data

    def serialize(self):
        raise NotImplementedError



from binascii import hexlify, unhexlify

from binary.unsigned_integer.writer import write_bit8

from crypto.transactions.serializers.base import BaseSerializer


class DelegateRegistrationSerializer(BaseSerializer):
    """Serializer handling delegate registration data
    """

    def serialize(self):
        delegate_bytes = hexlify(self.transaction['asset']['delegate']['username'].encode())
        self.bytes_data += write_bit8(len(delegate_bytes) // 2)
        self.bytes_data += unhexlify(delegate_bytes)
        return self.bytes_data



from crypto.configuration.fee import get_fee
from crypto.identity.public_key import PublicKey
from crypto.transactions.transaction import Transaction
from crypto.utils.message import Message


class BaseTransactionBuilder(object):

    def __init__(self):
        self.transaction = Transaction()
        self.transaction.type = getattr(self, 'transaction_type', None)
        self.transaction.fee = get_fee(getattr(self, 'transaction_type', None))

    def to_dict(self):
        return self.transaction.to_dict()

    def to_json(self):
        return self.transaction.to_json()

    def sign(self, passphrase):
        """Sign the transaction using the given passphrase

        Args:
            passphrase (str): passphrase associated with the account sending this transaction
        """
        self.transaction.senderPublicKey = PublicKey.from_passphrase(passphrase)
        message = Message.sign(self.transaction.to_bytes(), passphrase)
        self.transaction.signature = message.signature
        self.transaction.id = self.transaction.get_id()

    def second_sign(self, passphrase):
        """Sign the transaction using the given second passphrase

        Args:
            passphrase (str): 2nd passphrase associated with the account sending this transaction
        """
        message = Message.sign(self.transaction.to_bytes(skip_signature=False), passphrase)
        self.transaction.signSignature = message.signature
        self.transaction.id = self.transaction.get_id()

    def verify(self):
        self.transaction.verify()

    def second_verify(self):
        self.transaction.second_verify()



from binascii import hexlify, unhexlify

from coincurve import PublicKey as PubKey

from crypto.identity.private_key import PrivateKey


class PublicKey(object):

    def __init__(self, public_key):
        public_key = unhexlify(public_key.encode())
        self.public_key = PubKey(public_key)

    def to_hex(self):
        return hexlify(self.public_key.format()).decode()

    @classmethod
    def from_passphrase(cls, passphrase):
        private_key = PrivateKey.from_passphrase(passphrase)
        return private_key.public_key

    @classmethod
    def from_hex(cls, public_key):
        return cls(public_key)



from binascii import hexlify, unhexlify

from base58 import b58encode_check

from binary.unsigned_integer.reader import read_bit32, read_bit64

from crypto.transactions.deserializers.base import BaseDeserializer


class TransferDeserializer(BaseDeserializer):

    def deserialize(self):
        starting_position = int(self.asset_offset / 2)

        self.transaction.amount = read_bit64(self.serialized, offset=starting_position)
        self.transaction.expiration = read_bit32(self.serialized, offset=starting_position + 8)

        recipient_start_index = (int(self.asset_offset / 2) + 12) * 2
        recipientId = hexlify(self.serialized)[recipient_start_index:recipient_start_index + 42]
        self.transaction.recipientId = b58encode_check(unhexlify(recipientId)).decode()

        self.transaction.parse_signatures(
            hexlify(self.serialized),
            self.asset_offset + (8 + 4 + 21) * 2
        )
        return self.transaction



class BaseDeserializer(object):

    serialized = None
    asset_offset = None
    transaction = None

    def __init__(self, serialized, asset_offset, transaction):
        self.serialized = serialized
        self.asset_offset = asset_offset
        self.transaction = transaction

    def deserialize(self):
        raise NotImplementedError



import inspect
from binascii import hexlify, unhexlify
from hashlib import sha256
from importlib import import_module

from binary.unsigned_integer.reader import read_bit32, read_bit64, read_bit8

from crypto.constants import (
    TRANSACTION_MULTI_SIGNATURE_REGISTRATION, TRANSACTION_SECOND_SIGNATURE_REGISTRATION,
    TRANSACTION_TYPES, TRANSACTION_VOTE
)
from crypto.identity.address import address_from_public_key
from crypto.transactions.deserializers.base import BaseDeserializer


class Deserializer(object):

    serialized = None

    def __init__(self, serialized):
        self.serialized = unhexlify(serialized)

    def deserialize(self):
        """Deserialize transaction

        Returns:
            object: returns Transaction resource object
        """
        # circular import with transaction.py :( - I'm thinking of just returning a dict here
        # which then needs to be passed to a Transaction object, instead of returning a Transaction
        # object
        from crypto.transactions.transaction import Transaction

        transaction = Transaction()
        transaction.version = read_bit8(self.serialized, offset=1)
        transaction.network = read_bit8(self.serialized, offset=2)
        transaction.type = read_bit8(self.serialized, offset=3)
        transaction.timestamp = read_bit32(self.serialized, offset=4)
        transaction.senderPublicKey = hexlify(self.serialized)[16:66+16].decode()
        transaction.fee = read_bit64(self.serialized, offset=41)

        vendor_field_length = read_bit8(self.serialized, offset=49)
        if vendor_field_length > 0:
            vendor_field_offset = (49 + 8) * 2
            vendorField_take = vendor_field_length * 2
            transaction.vendorFieldHex = hexlify(
                self.serialized
            )[vendor_field_offset:vendorField_take]

        asset_offset = (49 + 1) * 2 + vendor_field_length * 2

        handled_tranasction = self._handle_transaction_type(asset_offset, transaction)
        transaction.amount = handled_tranasction.amount
        transaction.version = handled_tranasction.version
        if transaction.version == 1:
            transaction = self._handle_version_one(transaction)
        elif transaction.version == 2:
            transaction = self._handle_version_two(transaction)
        else:
            raise Exception('should this ever happen?')  # todo: do we need this?

        return transaction

    def _handle_transaction_type(self, asset_offset, transaction):
        """Handle deserialization for a given transaction type

        Args:
            asset_offset (int):
            transaction (object): Transaction resource object

        Returns:
            object: Transaction resource object of currently deserialized data
        """
        deserializer_name = TRANSACTION_TYPES[transaction.type]
        module = import_module('crypto.transactions.deserializers.{}'.format(deserializer_name))
        for attr in dir(module):
            # If attr name is `BaseDeserializer`, skip it as it's a class and also has a
            # subclass of BaseDeserializer
            if attr == 'BaseDeserializer':
                continue

            attribute = getattr(module, attr)
            if inspect.isclass(attribute) and issubclass(attribute, BaseDeserializer):
                # this attribute is actually a specific deserializer that we want to use
                deserializer = attribute
                break
        return deserializer(self.serialized, asset_offset, transaction).deserialize()

    def _handle_version_one(self, transaction):
        """Handle deserialization for version one

        Args:
            transaction (object): Transaction resource object

        Returns:
            object: Transaction resource object of currently deserialized data
        """
        if transaction.secondSignature:
            transaction.secondSignature = transaction.secondSignature

        if transaction.type is TRANSACTION_VOTE:
            transaction.recipientId = address_from_public_key(
                transaction.senderPublicKey, transaction.network
            )

        if transaction.type is TRANSACTION_MULTI_SIGNATURE_REGISTRATION:
            transaction.asset['multisignature']['keysgroup'] = [
                '+{}'.format(key) for key in transaction.asset['multisignature']['keysgroup']
            ]

        if transaction.vendorFieldHex:
            transaction.vendorField = unhexlify(transaction.vendorFieldHex)

        if not transaction.id:
            transaction.id = transaction.get_id()

        if transaction.type is TRANSACTION_SECOND_SIGNATURE_REGISTRATION:
            transaction.recipientId = address_from_public_key(
                transaction.senderPublicKey, transaction.network
            )

        if transaction.type is TRANSACTION_MULTI_SIGNATURE_REGISTRATION:
            transaction.recipientId = address_from_public_key(
                transaction.senderPublicKey, transaction.network
            )

        return transaction

    def _handle_version_two(self, transaction):
        """Handle deserialization for version two

        Args:
            transaction (object): Transaction resource object

        Returns:
            object: Transaction resource object of currently deserialized data
        """
        transaction.id = sha256(transaction)  # todo serialize



from crypto.constants import TRANSACTION_SECOND_SIGNATURE_REGISTRATION
from crypto.identity.public_key import PublicKey
from crypto.transactions.builder.base import BaseTransactionBuilder


class SecondSignatureRegistration(BaseTransactionBuilder):

    transaction_type = TRANSACTION_SECOND_SIGNATURE_REGISTRATION

    def __init__(self, second_passphrase, fee=None):
        """Create a second signature registration transaction

        Args:
            second_passphrase (str): used for signing a transaction the 2nd time
            fee (int, optional): fee used for the transaction (default is already set)
        """
        super().__init__()
        public_key = PublicKey.from_passphrase(second_passphrase)
        self.transaction.asset['signature'] = {'publicKey': public_key}
        if fee:
            self.transaction.fee = fee



from binascii import unhexlify

from binary.unsigned_integer.writer import write_bit8

from crypto.transactions.serializers.base import BaseSerializer


class IPFSSerializer(BaseSerializer):
    """Serializer handling ipfs data
    """

    def serialize(self):
        dag = self.transaction['asset']['ipfs']['dag']
        self.bytes_data += write_bit8(len(dag) // 2)
        self.bytes_data += unhexlify(dag)
        return self.bytes_data



import inspect
from binascii import hexlify, unhexlify
from importlib import import_module

from binary.hex.writer import write_high, write_low
from binary.unsigned_integer.writer import write_bit32, write_bit64, write_bit8

from crypto.configuration.network import get_network
from crypto.constants import TRANSACTION_TYPES
from crypto.exceptions import ArkSerializerException
from crypto.transactions.serializers.base import BaseSerializer


class Serializer(object):

    transaction = None

    def __init__(self, transaction):
        if not transaction:
            raise ArkSerializerException('No transaction data provided')
        self.transaction = transaction

    def serialize(self):
        """Perform AIP11 compliant serialization

        Returns:
            bytes: bytes string
        """
        network_config = get_network()
        bytes_data = bytes()
        bytes_data += write_bit8(0xff)
        bytes_data += write_low(self.transaction.get('version') or 0x01)
        bytes_data += write_bit8(self.transaction.get('network') or network_config['version'])
        bytes_data += write_low(self.transaction['type'])
        bytes_data += write_bit32(self.transaction['timestamp'])
        bytes_data += write_high(self.transaction['senderPublicKey'])
        bytes_data += write_bit64(self.transaction['fee'])

        if self.transaction.get('vendorField'):
            vendorFieldLength = len(self.transaction['vendorField'])
            bytes_data += write_bit8(vendorFieldLength)
            bytes_data += self.transaction['vendorField']
        elif self.transaction.get('vendorFieldHex'):
            vendorField_hex_length = len(self.transaction['vendorFieldHex'])
            bytes_data += write_bit8(vendorField_hex_length / 2)
            bytes_data += self.transaction['vendorFieldHex']
        else:
            bytes_data += write_bit8(0x00)

        bytes_data = self._handle_transaction_type(bytes_data)
        bytes_data = self._handle_signature(bytes_data)

        return hexlify(bytes_data).decode()

    def _handle_transaction_type(self, bytes_data):
        """Serialize transaction specific data (eg. delegate registration)

        Args:
            bytes_data (bytes): already serialized data about a transaction (eg. version, network)

        Returns:
            bytes: bytes string
        """
        serializer_name = TRANSACTION_TYPES[self.transaction['type']]

        module = import_module('crypto.transactions.serializers.{}'.format(serializer_name))
        for attr in dir(module):
            # If attr name is `BaseSerializer`, skip it as it's a class and also has a
            # subclass of BaseSerializer
            if attr == 'BaseSerializer':
                continue

            attribute = getattr(module, attr)
            if inspect.isclass(attribute) and issubclass(attribute, BaseSerializer):
                # this attribute is actually a specific serializer that we want to use
                serializer = attribute
                break
        return serializer(self.transaction, bytes_data).serialize()

    def _handle_signature(self, bytes_data):
        """Serialize signature data of the transaction

        Args:
            bytes_data (bytes): already serialized data

        Returns:
            bytes: bytes string
        """
        if self.transaction.get('signature'):
            bytes_data += unhexlify(self.transaction['signature'])

        if self.transaction.get('secondSignature'):
            bytes_data += unhexlify(self.transaction['secondSignature'])
        elif self.transaction.get('signSignature'):
            bytes_data += unhexlify(self.transaction['signSignature'])

        if self.transaction.get('signatures'):
            bytes_data += write_bit8(0xff)
            bytes_data += unhexlify(''.join(self.transaction['signatures']))

        return bytes_data



from binascii import hexlify

from binary.unsigned_integer.reader import read_bit8

from crypto.transactions.deserializers.base import BaseDeserializer


class VoteDeserializer(BaseDeserializer):

    def deserialize(self):
        starting_position = int(self.asset_offset / 2)

        vote_length = read_bit8(self.serialized, starting_position) & 0xff

        self.transaction.asset['votes'] = []

        for index in range(vote_length):
            starting_position = self.asset_offset + 2 + (index * 2 * 34)
            vote = hexlify(self.serialized)[starting_position:starting_position + 2 * 34].decode()
            if vote[1] == '1':
                operator = '+'
            else:
                operator = '-'
            vote = '{}{}'.format(operator, vote[2:])

            self.transaction.asset['votes'].append(vote)

        self.transaction.parse_signatures(
            hexlify(self.serialized),
            self.asset_offset + 2 + (vote_length * 34 * 2)
        )
        return self.transaction



from binascii import hexlify

from base58 import b58decode_check

from binary.hex.writer import write_high
from binary.unsigned_integer.writer import write_bit32, write_bit64

from crypto.transactions.serializers.base import BaseSerializer


class MultiPaymentSerializer(BaseSerializer):
    """Serializer handling multi payment data
    """

    def serialize(self):
        self.bytes_data += write_bit32(len(self.transaction['asset']['payments']))
        for payment in self.transaction['asset']['payments']:
            self.bytes_data += write_bit64(payment['amount'])
            recipientId = b58decode_check(hexlify(payment['recipientId']).encode())
            self.bytes_data += write_high(recipientId)
        return self.bytes_data



from crypto.constants import TRANSACTION_FEES

fees = TRANSACTION_FEES.copy()


def get_fee(transaction_type):
    """Get a fee for a given transaction type

    Args:
        transaction_type (int): transaction type for which we wish to get a fee

    Returns:
        int: transaction fee
    """
    return fees.get(transaction_type)


def set_fee(transaction_type, value):
    """Set a fee

    Args:
        transaction_type (int): transaction_type for which we wish to set a fee
        value (int): fee for a given transaction type
    """
    global fees
    fees[transaction_type] = value



from crypto.networks.devnet import Devnet

network = {}


def set_network(network_object):
    """Set what network you want to use in the crypto library

    Args:
        network_object (Network object): Testnet, Devnet, Mainnet
    """
    global network
    network = {
        'epoch': network_object.epoch,
        'version': network_object.version,
        'wif': network_object.wif,
    }


def get_network():
    """Get settings for a selected network, default network is devnet

    Returns:
        dict: network settings (default network is devnet)
    """
    if not network:
        set_network(Devnet)
    return network


def set_custom_network(epoch, version, wif):
    """Set custom network

    Args:
        epoch (datetime): chains epoch time
        version (int): chains version
        wif (int): chains wif
    """
    global network
    network = {
        'epoch': epoch,
        'version': version,
        'wif': wif,
    }



from crypto.constants import TRANSACTION_DELEGATE_REGISTRATION
from crypto.identity.public_key import PublicKey
from crypto.transactions.builder.base import BaseTransactionBuilder


class DelegateRegistration(BaseTransactionBuilder):

    transaction_type = TRANSACTION_DELEGATE_REGISTRATION

    def __init__(self, username, fee=None):
        """Create a delegate registration transaction

        Args:
            username (str): username of a delegate you want to register
            fee (int, optional): fee used for the transaction (default is already set)
        """
        super().__init__()
        self.transaction.asset['delegate'] = {'username': username}
        if fee:
            self.transaction.fee = fee

    def sign(self, passphrase):
        public_key = PublicKey.from_passphrase(passphrase)
        self.transaction.asset['delegate']['publicKey'] = public_key
        super().sign(passphrase)



from crypto.constants import TRANSACTION_DELEGATE_RESIGNATION
from crypto.transactions.builder.base import BaseTransactionBuilder


class DelegateResignation(BaseTransactionBuilder):

    transaction_type = TRANSACTION_DELEGATE_RESIGNATION

    def __init__(self, fee=None):
        """Create a delegate resignation transaction

        Args:
            fee (int, optional): fee used for the transaction (default is already set)
        """
        super().__init__()
        if fee:
            self.transaction.fee = fee



from crypto.constants import TRANSACTION_MULTI_SIGNATURE_REGISTRATION
from crypto.transactions.builder.base import BaseTransactionBuilder


class MultiSignatureRegistration(BaseTransactionBuilder):

    transaction_type = TRANSACTION_MULTI_SIGNATURE_REGISTRATION

    def __init__(self, min_signatures, lifetime, keysgroup, fee=None):
        """Create a new multi signature transaction

        Args:
            min_signatures (int): minimum required signatures
            lifetime (int): transaction lifetime
            keysgroup (list): list of signatures required
            fee (int, optional): fee used for the transaction (default is already set)
        """
        super().__init__()
        self.transaction.asset = {
            'multisignature': {
                'min': min_signatures,
                'lifetime': lifetime,
                'keysgroup': keysgroup,
            },
        }
        # default transaction fee for this type i set in the base builder
        transaction_fee = fee or self.transaction.fee
        self.transaction.fee = (len(keysgroup) + 1) * transaction_fee



from binascii import hexlify

from crypto.transactions.deserializers.base import BaseDeserializer


class SecondSignatureRegistrationDeserializer(BaseDeserializer):

    def deserialize(self):
        starting_position = int(self.asset_offset)
        public_key = hexlify(self.serialized)[starting_position:starting_position + 66]
        self.transaction.asset = {
            'signature': {
                'publicKey': public_key.decode()
            }
        }

        self.transaction.parse_signatures(hexlify(self.serialized), self.asset_offset + 66)
        return self.transaction



class ArkCryptoException(Exception):
    pass


class ArkSerializerException(ArkCryptoException):
    """Raised when there's a serializer related issue"""


class ArkInvalidTransaction(ArkCryptoException):
    """Raised when transaction is not valid"""



import sys

import setuptools


requires = [
    'base58',
    'binary-helpers',
    'coincurve'
]

tests_require = [
    'flake8>=3.5.0',
    'flake8-import-order>=0.17.1',
    'flake8-print>=3.1.0',
    'flake8-quotes>=1.0.0',
    'pytest>=3.6.1',
    'pytest-cov>=2.5.1'
]

extras_require = {
    'test': tests_require,
    'dev': requires + tests_require
}

setup_requires = ['pytest-runner'] if {'pytest', 'test', 'ptr'}.intersection(sys.argv) else []

setuptools.setup(
    name='arkecosystem-crypto',
    description='A simple Cryptography Implementation in Python for the ARK Blockchain.',
    version='0.1.1',
    author='Ark Ecosystem',
    author_email='info@ark.io',
    url='https://github.com/ArkEcosystem/python-crypto',
    packages=setuptools.find_packages(exclude=['tests', 'tests.*']),
    install_requires=requires,
    extras_require=extras_require,
    tests_require=tests_require,
    setup_requires=setup_requires,
)



from binascii import hexlify

from base58 import b58decode_check

from binary.hex.writer import write_high
from binary.unsigned_integer.writer import write_bit32, write_bit64

from crypto.transactions.serializers.base import BaseSerializer


class TransferSerializer(BaseSerializer):
    """Serializer handling transfer data
    """

    def serialize(self):
        self.bytes_data += write_bit64(self.transaction['amount'])
        self.bytes_data += write_bit32(self.transaction.get('expiration', 0))
        recipientId = hexlify(b58decode_check(self.transaction['recipientId']))
        self.bytes_data += write_high(recipientId)
        return self.bytes_data



TRANSACTION_TRANSFER = 0
TRANSACTION_SECOND_SIGNATURE_REGISTRATION = 1
TRANSACTION_DELEGATE_REGISTRATION = 2
TRANSACTION_VOTE = 3
TRANSACTION_MULTI_SIGNATURE_REGISTRATION = 4
TRANSACTION_IPFS = 5
TRANSACTION_TIMELOCK_TRANSFER = 6
TRANSACTION_MULTI_PAYMENT = 7
TRANSACTION_DELEGATE_RESIGNATION = 8

TRANSACTION_TYPES = {
    TRANSACTION_TRANSFER: 'transfer',
    TRANSACTION_SECOND_SIGNATURE_REGISTRATION: 'second_signature_registration',
    TRANSACTION_DELEGATE_REGISTRATION: 'delegate_registration',
    TRANSACTION_VOTE: 'vote',
    TRANSACTION_MULTI_SIGNATURE_REGISTRATION: 'multi_signature_registration',
    TRANSACTION_IPFS: 'ipfs',
    TRANSACTION_TIMELOCK_TRANSFER: 'timelock_transfer',
    TRANSACTION_MULTI_PAYMENT: 'multi_payment',
    TRANSACTION_DELEGATE_RESIGNATION: 'delegate_resignation',
}

TRANSACTION_FEES = {
    TRANSACTION_TRANSFER: 10000000,
    TRANSACTION_SECOND_SIGNATURE_REGISTRATION: 500000000,
    TRANSACTION_DELEGATE_REGISTRATION: 2500000000,
    TRANSACTION_VOTE: 100000000,
    TRANSACTION_MULTI_SIGNATURE_REGISTRATION: 500000000,
    TRANSACTION_IPFS: 0,
    TRANSACTION_TIMELOCK_TRANSFER: 0,
    TRANSACTION_MULTI_PAYMENT: 0,
    TRANSACTION_DELEGATE_RESIGNATION: 0,
}



from binascii import unhexlify

from binary.unsigned_integer.writer import write_bit8

from crypto.transactions.serializers.base import BaseSerializer


class DelegateRegistrationSerializer(BaseSerializer):
    """Serializer handling delegate registration data
    """

    def serialize(self):
        keysgroup = []
        if self.transaction.get('version') is None or self.transaction['version'] == 1:
            for key in self.transaction['asset']['multisignature']['keysgroup']:
                keysgroup.append(key[1::] if key.startswith('+') else key)
        else:
            keysgroup = self.transaction['asset']['multisignature']['keysgroup']
        self.bytes_data += write_bit8(self.transaction['asset']['multisignature']['min'])
        self.bytes_data += write_bit8(len(self.transaction['asset']['multisignature']['keysgroup']))
        self.bytes_data += write_bit8(self.transaction['asset']['multisignature']['lifetime'])
        self.bytes_data += unhexlify(''.join(keysgroup))
        return self.bytes_data



import json
from binascii import unhexlify
from hashlib import sha256
from struct import pack

from base58 import b58decode_check

from binary.hex.writer import write_high
from binary.unsigned_integer.writer import write_bit32, write_bit64, write_bit8

from crypto.constants import (
    TRANSACTION_DELEGATE_REGISTRATION, TRANSACTION_MULTI_SIGNATURE_REGISTRATION,
    TRANSACTION_SECOND_SIGNATURE_REGISTRATION, TRANSACTION_VOTE
)
from crypto.exceptions import ArkInvalidTransaction
from crypto.transactions.deserializer import Deserializer
from crypto.transactions.serializer import Serializer
from crypto.utils.message import Message
from crypto.utils.slot import get_time


TRANSACTION_ATTRIBUTES = {
    'amount': 0,
    'asset': dict,
    'fee': None,
    'id': None,
    'network': None,
    'recipientId': None,
    'secondSignature': None,
    'senderPublicKey': None,
    'signature': None,
    'signatures': None,
    'signSignature': None,
    'timestamp': get_time,
    'type': None,
    'vendorField': None,
    'vendorFieldHex': None,
    'version': None,
}


class Transaction(object):

    def __init__(self, *args, **kwargs):
        for attribute, attribute_value in TRANSACTION_ATTRIBUTES.items():
            if callable(attribute_value):
                attribute_value = attribute_value()
            if attribute in kwargs:
                attribute_value = kwargs[attribute]
            setattr(self, attribute, attribute_value)

    def get_id(self):
        """Convert the byte representation to a unique identifier

        Returns:
            str:
        """
        return sha256(self.to_bytes(False, False)).hexdigest()

    def to_dict(self):
        """Convert the transaction into a dictionary representation

        Returns:
            dict: only includes values that are set
        """
        data = {}
        for key in TRANSACTION_ATTRIBUTES.keys():
            attribute = getattr(self, key, None)
            if attribute is None:
                continue
            # todo: get rid of the bytes check and handle this outside of the to_dict function
            data[key] = attribute.decode() if isinstance(attribute, bytes) else attribute
        return data

    def to_json(self):
        data = self.to_dict()
        return json.dumps(data)

    def to_bytes(self, skip_signature=True, skip_second_signature=True):
        """Convert the transaction to its byte representation

        Args:
            skip_signature (bool, optional): do you want to skip the signature
            skip_second_signature (bool, optional): do you want to skip the 2nd signature

        Returns:
            bytes: bytes representation of the transaction
        """
        bytes_data = bytes()
        bytes_data += write_bit8(self.type)
        bytes_data += write_bit32(self.timestamp)
        bytes_data += write_high(self.senderPublicKey)

        if self.recipientId:
            bytes_data += b58decode_check(self.recipientId)
        else:
            bytes_data += pack('21x')

        if self.vendorField and len(self.vendorField) < 64:
            bytes_data += self.vendorField
            bytes_data += pack('{}x'.format(64 - len(self.vendorField)))
        else:
            bytes_data += pack('64x')

        bytes_data += write_bit64(self.amount)
        bytes_data += write_bit64(self.fee)

        bytes_data = self._handle_transaction_type(bytes_data)
        bytes_data = self._handle_signature(bytes_data, skip_signature, skip_second_signature)

        return bytes_data

    def parse_signatures(self, serialized, start_offset):
        """Parse the signature, second signature and multi signatures

        Args:
            serialized (str): parses a given serialized string
            start_offset (int):

        Returns:
            None: methods returns nothing
        """

        self.signature = serialized[start_offset:]
        multi_signature_offset = 0

        if not len(self.signature):
            self.signature = None
            return

        signature_length = int(self.signature[2:4], base=16) + 2
        self.signature = serialized[start_offset: start_offset + (signature_length * 2)]
        multi_signature_offset += signature_length * 2
        self.signSignature = serialized[start_offset + (signature_length * 2):]

        if not self.signSignature:
            self.signSignature = None
        elif 'ff' == self.signSignature[:2]:
            self.signSignature = None
        else:
            secondSignature_length = int(self.signSignature[2:4], base=16)
            self.signSignature = self.signSignature[:secondSignature_length * 2]
            multi_signature_offset += secondSignature_length * 2

        signatures = serialized[:start_offset + multi_signature_offset]

        if not signatures:
            return

        if 'ff' != signatures[:2]:
            return

        signatures = signatures[2:]
        self.signatures = []

        while True:
            mlength = int(signatures[2:4], base=16)
            if mlength > 0:
                self.signatures.append(signatures[:(mlength + 2) * 2])
            else:
                break

            signatures = signatures[(mlength + 2) * 2:]
            if not signatures:
                break

    def serialize(self):
        data = self.to_dict()
        return Serializer(data).serialize()

    def deserialize(self, serialized):
        return Deserializer(serialized).deserialize()

    def verify(self):
        """Verify the transaction. Method will raise an exception if invalid, if it's valid nothing
        will happen.
        """
        transaction = self.to_bytes()
        message = Message(transaction, self.signature, self.senderPublicKey)
        is_valid = message.verify()
        if not is_valid:
            raise ArkInvalidTransaction('Transaction could not be verified')

    def second_verify(self, passphrase):
        """Verify the transaction using the 2nd passphrase

        Args:
            passphrase (str): 2nd passphrase associated with the account sending this transaction
        """
        transaction = sha256(self.to_bytes()).digest()
        message = Message(transaction, self.signSignature, self.senderPublicKey)
        is_valid = message.verify()
        if not is_valid:
            raise ArkInvalidTransaction('Transaction could not be verified')

    def _handle_transaction_type(self, bytes_data):
        """Handled each transaction type differently

        Args:
            bytes_data (bytes): input the bytes data to which you want to append new bytes

        Raises:
            NotImplementedError: raised only if the child transaction doesn't implement this
            required method
        """
        if self.type == TRANSACTION_SECOND_SIGNATURE_REGISTRATION:
            public_key = self.asset['signature']['publicKey']
            bytes_data += unhexlify(public_key)
        elif self.type == TRANSACTION_DELEGATE_REGISTRATION:
            bytes_data += self.asset['delegate']['username'].encode()
        elif self.type == TRANSACTION_VOTE:
            bytes_data += ''.join(self.asset['votes']).encode()
        elif self.type == TRANSACTION_MULTI_SIGNATURE_REGISTRATION:
            bytes_data += write_bit8(self.asset['multisignature']['min'])
            bytes_data += write_bit8(self.asset['multisignature']['lifetime'])
            bytes_data += ''.join(self.asset['multisignature']['keysgroup']).encode()
        return bytes_data

    def _handle_signature(self, bytes_data, skip_signature, skip_second_signature):
        """Internal method, used to handle the signature

        Args:
            bytes_data (bytes): input the bytes data to which you want to append new bytes from
            signature
            skip_signature (bool): whether you want to skip it or not
            skip_second_signature (bool): whether you want to skip it or not

        Returns:
            bytes: bytes string
        """
        if not skip_signature and self.signature:
            bytes_data += write_high(self.signature)
        if not skip_second_signature and self.signSignature:
            bytes_data += write_high(self.signSignature)
        return bytes_data



import json
from binascii import unhexlify

from crypto.identity.private_key import PrivateKey
from crypto.identity.public_key import PublicKey


class Message(object):

    def __init__(self, message, signature, public_key):
        self.public_key = public_key
        self.signature = signature
        self.message = message

    @classmethod
    def sign(cls, message, passphrase):
        """Signs a message

        Args:
            message (str/bytes): a message you wish to sign
            passphrase (str/byes): passphrase you wish to use to sign the message

        Returns:
            Message: returns a message object
        """
        message_byes = message if isinstance(message, bytes) else message.encode()
        passphrase = passphrase.decode() if isinstance(passphrase, bytes) else passphrase
        private_key = PrivateKey.from_passphrase(passphrase)
        signature = private_key.sign(message_byes)
        return cls(message, signature, private_key.public_key)

    def verify(self):
        """Verify the Message object

        Returns:
            bool: returns a boolean - true if verified, false if not
        """
        message = self.message if isinstance(self.message, bytes) else self.message.encode()
        key = PublicKey.from_hex(self.public_key)
        signature = unhexlify(self.signature)
        is_verified = key.public_key.verify(signature, message)
        return is_verified

    def to_dict(self):
        """Return a dictionary of the message

        Returns:
            dict: dictionary consiting of public_key, signature and message
        """
        data = {
            'public_key': self.public_key,
            'signature': self.signature,
            'message': self.message,
        }
        return data

    def to_json(self):
        """Returns a json string of the the message

        Returns:
            str: json string consisting of public_key, signature and message
        """
        data = self.to_dict()
        return json.dumps(data)



from binascii import hexlify, unhexlify

from binary.unsigned_integer.reader import read_bit8

from crypto.transactions.deserializers.base import BaseDeserializer


class DelegateRegistrationDeserializer(BaseDeserializer):

    def deserialize(self):
        starting_position = int(self.asset_offset / 2)

        username_length = read_bit8(self.serialized, starting_position) & 0xff
        start_index = self.asset_offset + 2
        end_index = start_index + (username_length * 2)
        username = hexlify(self.serialized)[start_index:end_index]
        username = unhexlify(username)

        self.transaction.asset['delegate'] = {'username': username.decode()}

        self.transaction.parse_signatures(
            hexlify(self.serialized),
            self.asset_offset + (username_length + 1) * 2
        )
        return self.transaction



import hashlib
from binascii import unhexlify

from base58 import b58decode_check, b58encode_check

from binary.unsigned_integer.writer import write_bit8

from crypto.configuration.network import get_network
from crypto.identity.private_key import PrivateKey


def address_from_public_key(public_key, network_version=None):
    """Get an address from a public key

    Args:
        public_key (str):
        network_version (int, optional):

    Returns:
        bytes:
    """
    if not network_version:
        network = get_network()
        network_version = network['version']

    ripemd160 = hashlib.new('ripemd160', unhexlify(public_key.encode()))
    seed = write_bit8(network_version) + ripemd160.digest()
    return b58encode_check(seed).decode()


def address_from_private_key(private_key, network_version=None):
    """Get an address from private key

    Args:
        private_key (string):
        network_version (int, optional):

    Returns:
        TYPE: Description
    """
    if not network_version:
        network = get_network()
        network_version = network['version']

    private_key = PrivateKey.from_hex(private_key)
    ripemd160 = hashlib.new('ripemd160', unhexlify(private_key.public_key))
    seed = write_bit8(network_version) + ripemd160.digest()
    return b58encode_check(seed).decode()


def address_from_passphrase(passphrase, network_version=None):
    """Get an address from passphrase

    Args:
        passphrase (str):
        network_version (int, optional):

    Returns:
        string: address
    """
    if not network_version:
        network = get_network()
        network_version = network['version']

    private_key = hashlib.sha256(passphrase.encode()).hexdigest()
    address = address_from_private_key(private_key)
    return address


def validate_address(address, network_version=None):
    """Validate a given address

    Args:
        address (str): address you wish to validate
        network_version (None, optional): integer, version of the network

    Returns:
        bool:
    """
    if not network_version:
        network = get_network()
        network_version = network['version']

    return network_version == b58decode_check(address)[0]
